from enum import Enum, auto

from . import utils

class Type(Enum):
    Point       = auto()
    Interval    = auto()

class Event(utils.AutoName):
    Validate    = auto()    # interval to be used for running glassesValidator
    Sync_Camera = auto()    # point to be used for synchronizing different cameras
    Sync_ET_Data= auto()    # episode to be used for synchronization of eye tracker data to scene camera (e.g. using VOR)
    Trial       = auto()    # episode for which to map gaze to plane(s): output for files to be provided to user
events = [x.value for x in Event]
utils.register_type(utils.CustomTypeEntry(Event,'__enum.Event__', utils.enum_val_2_str, lambda x: getattr(Event, x.split('.')[1])))

type_map = {
    Event.Validate    : Type.Interval,
    Event.Sync_Camera : Type.Point,
    Event.Sync_ET_Data: Type.Interval,
    Event.Trial       : Type.Interval,
}

tooltip_map = {
    Event.Validate    : 'Validation episode',
    Event.Sync_Camera : 'Camera sync point',
    Event.Sync_ET_Data: 'Eye tracker synchronization episode',
    Event.Trial       : 'Trial episode',
}

def flatten_annotation_dict(annotations: dict[Event, list[list[int]]]) -> dict[Event, list[int]]:
    annotations_flat: dict[Event, list[int]] = {}
    for e in Event:  # iterate over this for consistent ordering
        if e not in annotations:
            continue
        if annotations[e] and isinstance(annotations[e][0],list):
            annotations_flat[e] = [i for iv in annotations[e] for i in iv]
        else:
            annotations_flat[e] = annotations[e].copy()
    return annotations_flat


# -*- coding: utf-8 -*-
from psychopy import visual, monitors, core, event, tools
import numpy as np
import pandas as pd
import json
import traceback
import random
from collections import defaultdict
import cv2

class ABCFixPoint:
    def __init__(self, win, outer_diameter=0.7, inner_diameter=0.1,
                 outer_color='black', inner_color='white', units='degFlatPos'):
        self.outer_dot      = visual.Circle(win, fillColor = outer_color,
                                            radius = outer_diameter/2,
                                            units = units)
        self.inner_dot      = visual.Circle(win, fillColor = outer_color,
                                            radius = inner_diameter/2,
                                            units = units)
        self.line_vertical  = visual.Rect(win, fillColor=inner_color,
                                          width=inner_diameter, height=outer_diameter,
                                          units = units)
        self.line_horizontal= visual.Rect(win, fillColor=inner_color,
                                          width=outer_diameter, height=inner_diameter,
                                          units = units)

    def set_size(self, size: float):
        self.outer_dot.radius       = size / 2
        self.line_vertical.size     = (self.line_vertical.width, size)
        self.line_horizontal.size   = (size, self.line_horizontal.height)

    def set_pos(self, pos):
        self.outer_dot.pos          = pos
        self.inner_dot.pos          = pos
        self.line_vertical.pos      = pos
        self.line_horizontal.pos    = pos

    def get_pos(self):
        return self.outer_dot.pos

    def get_size(self):
        return self.outer_dot.size

    def draw(self):
        self.outer_dot.draw()
        self.line_vertical.draw()
        self.line_horizontal.draw()
        self.inner_dot.draw()

class SegmentationMarker:
    def __init__(self, win: visual.Window, refresh_rate: int, duration: float,
                 marker_size: float, units: str, margin: float, border_bits: int, background_color: str):
        self.win = win
        self.refresh_rate = refresh_rate
        self.duration = duration
        self.size = marker_size
        self.units = units
        self.margin = margin
        self.border_bits = border_bits
        self.background_color = background_color

        self._marker    = visual.ImageStim(win, image=None, size=self.size, units=self.units)
        self._marker_bg = visual.Rect(win,
                                      units =self.units,
                                      width =self.size*(1+self.margin*2),
                                      height=self.size*(1+self.margin*2),
                                      fillColor=self.background_color)

    def draw(self, m_id: int):
        check_escape(self.win)
        self._marker.image = get_aruco_marker(m_id, self.size, self.units, self.border_bits, self.win)
        n_segment_marker_frames = int(self.duration*self.refresh_rate)
        for _ in range(n_segment_marker_frames):
            self._marker_bg.draw()
            self._marker.draw()
            self.win.flip()

def read_coord_file(file):
    return pd.read_csv(file, dtype=defaultdict(lambda: np.float32, ID='int32', color='str')).dropna(axis=0, how='all').set_index('ID')

_aruco_dict = cv2.aruco.getPredefinedDictionary(cv2.aruco.DICT_4X4_250)
def get_aruco_marker(m_id: int, size: float, units: str, border_bits: int, win: visual.Window):
    size = tools.monitorunittools.convertToPix(np.array([-.5*size,.5*size]),np.array([0.,0.]),units,win)
    size = int(size[1]-size[0])
    # NB: flipud because PsychoPy draws images loaded from memory upside down
    return np.flipud(cv2.aruco.generateImageMarker(_aruco_dict, m_id, size, borderBits=border_bits).astype(np.float32)/255.*2-1)

def check_escape(win: visual.Window):
    if 'escape' in event.getKeys():
        win.close()
        core.quit()


def prepare_validation(win: visual.Window, config: dict, screen_config: dict):
    # get markers and target positions
    target_positions   = read_coord_file(config["targets"]["file"])
    fiducial_positions = read_coord_file(config["markers"]["file"])
    # prepare target order
    targets: list[int] = []
    if config["targets"]["first_ID"] not in target_positions.index:
        raise ValueError(f'Target ID set in setup.json->targets->first_ID ({config["targets"]["first_ID"]}) is not in the targets file {config["targets"]["file"]}')
    ts = [t for t in target_positions.index.to_list() if t!=config["targets"]["first_ID"]]
    for _ in range(config["targets"]["n_repetitions"]):
        targets.append(config["targets"]["first_ID"])
        random.shuffle(ts)
        targets += ts

    # check there are no overlapping ArUco markers
    indicator_IDs = []
    for ri,_ in enumerate(config["markers"]["replace_IDs"]):
        off = config["markers"]["replace_ID_start"]+ri*config["markers"]["replace_ID_offset"]
        indicator_IDs += [i+off for i in target_positions.index]
    if duplicates:=set(indicator_IDs) & set(fiducial_positions.index):
        raise ValueError(f"For the current setup, marker IDs specified in marker file {config['markers']['file']} would also appear as target indicator markers. Adapt the marker positions file or config->markers->replace_ID_start, config->markers->replace_ID_offset and replace_IDs. Duplicates: {sorted(list(duplicates))}.")

    # Create a background with ArUco markers and circle placeholders where target will appear
    stimList = []
    # draw all ArUco markers
    # NB ArUco markers are drawn using the 'deg' and not the 'degFlatPos' coordinate system so
    # that these markers are evenly spaced and not more eccentric than necessary
    for m_id, marker in fiducial_positions.iterrows():
        aruco_im = visual.ImageStim(win, image=None, size=config["markers"]["size"], units=config["markers"]["units"])
        aruco_bg = visual.Rect(win, units=config["markers"]["units"],
                               width =config["markers"]["size"]*(1+config["markers"]["margin"]*2),
                               height=config["markers"]["size"]*(1+config["markers"]["margin"]*2),
                               fillColor=config["markers"]["background_color"])

        aruco_bg.pos = [marker.x, marker.y]
        aruco_im.image = get_aruco_marker(m_id, config["markers"]["size"], config["markers"]["units"], config["markers"]["border_bits"], win)
        aruco_im.pos = [marker.x, marker.y]

        stimList.append(aruco_bg)
        stimList.append(aruco_im)

        aruco_bg.draw()
        aruco_im.draw()

    # draw placeholder circles at all target positions
    if (use_file_color:='color' in target_positions.columns) or config["targets"]["placeholder_color"]:
        # if color specified in the the target_positions file, use that. Else, fall back on placeholder color
        # specified in config. If neither is specified, do not draw placeholders
        for _, target in target_positions.iterrows():
            circle = visual.Circle(win, radius=config["targets"]["placeholder"]["diameter"]/2, units=config["targets"]["units"], fillColor=target.color if use_file_color else config["targets"]["placeholder_color"])
            circle.pos = [target.x, target.y]
            stimList.append(circle)
            circle.draw()

    # Get screenshot of background, so that we can draw unchanging things at once
    background = visual.ImageStim(win, visual.BufferImageStim(win, stim=stimList).image)    # because https://github.com/psychopy/psychopy/issues/840
    if screen_config["origin"] is not None:
        background.pos = [-x for x in screen_config["origin"]]
    return {'background': background, 'target_positions': target_positions, 'fiducial_positions': fiducial_positions, 'targets': targets}

def show_validation(win: visual.Window, config: dict, refresh_rate: int, task_vars: dict):
    # prepare visual objects
    fixation_target = ABCFixPoint(win, outer_color=config["targets"]["look"]["outer_color"], inner_color=config["targets"]["look"]["inner_color"], units=config["targets"]["units"])
    aruco_ims = [visual.ImageStim(win, image=None, size=config["markers"]["size"], units=config["markers"]["units"]) for _ in config["markers"]["replace_IDs"]]
    if (have_cue:=config["targets"]["cue"]["color"] is not None):
        cue = visual.Circle(win, radius=config["targets"]["cue"]["diameter"]/2, units=config["targets"]["units"], fillColor=config["targets"]["cue"]["color"])

    # prepare task parameters
    n_shrink_frames = int(config["targets"]["shrink"]["duration"]*refresh_rate)
    shrink_sizes    = np.linspace(config["targets"]["look"]["diameter_max"]/2, config["targets"]["look"]["diameter_min"]/2, n_shrink_frames)
    n_capture_frames= int(config["targets"]["duration"]*refresh_rate)

    # show fixation target sequence
    old_pos = task_vars['target_positions'].loc[task_vars['targets'][0],['x','y']].tolist()
    for i in task_vars['targets']:
        check_escape(win)

        # Move target to new position
        pos   = task_vars['target_positions'].loc[i,['x','y']].tolist()
        d     = np.sqrt((old_pos[0]-pos[0])**2 + (old_pos[1]-pos[1])**2)
        old_pos_pix = tools.monitorunittools.convertToPix([0.,0.],old_pos,config["targets"]["units"],win)
        pos_pix     = tools.monitorunittools.convertToPix([0.,0.],    pos,config["targets"]["units"],win)
        d_pix = np.sqrt((old_pos_pix[0]-pos_pix[0])**2 + (old_pos_pix[1]-pos_pix[1])**2)
        # Target should move at constant speed regardless of distance to cover, duration contains time to move
        # over width of whole screen. Adjust time to proportion of screen width covered by current move
        move_duration = max(config["targets"]["move"]["min_duration"], config["targets"]["move"]["duration"]*d_pix/win.size[0])
        n_move_frames = int(move_duration*refresh_rate)
        if config["targets"]["move"]["move_with_acceleration"]:
            accel   = 0 if not d else d/(move_duration/2)**2    # solve x=.5*a*t^2 for a, use d/2 for x
            moveVec = [0.,0.] if not d else [(x-y)/d for x,y in zip(pos,old_pos)]
            def calc_pos(frac):
                if frac<.5:
                    return [p+m*.5*accel*(frac*move_duration)**2 for p,m in zip(old_pos,moveVec)]
                else:
                    # implement deceleration by accelerating from the other side in backward time
                    return [p-m*.5*accel*((1-frac)*move_duration)**2 for p,m in zip(pos,moveVec)]
            tar_pos = [calc_pos(x) for x in np.linspace(0., 1., n_move_frames)]
        else:
            x_tar_pos = np.linspace(old_pos[0], pos[0], n_move_frames)
            y_tar_pos = np.linspace(old_pos[1], pos[1], n_move_frames)
            tar_pos = [[x,y] for x,y in zip(x_tar_pos,y_tar_pos)]

        # two markers indicate which target is shown (two for redundancy)
        for ri,r in enumerate(config["markers"]["replace_IDs"]):
            m_id = config["markers"]["replace_ID_start"]+ri*config["markers"]["replace_ID_offset"]+i
            aruco_ims[ri].image = get_aruco_marker(m_id, config["markers"]["size"], config["markers"]["units"], config["markers"]["border_bits"], win)
            aruco_ims[ri].pos   = task_vars['fiducial_positions'].loc[r,['x','y']].tolist()

        # move to next target
        fixation_target.set_size(config["targets"]["move"]["diameter"])
        cue.pos = pos
        for p in tar_pos:
            task_vars['background'].draw()
            if have_cue:
                cue.draw()
            fixation_target.set_pos(p)
            fixation_target.draw()
            win.flip()

        # shrink
        for s in shrink_sizes:
            task_vars['background'].draw()
            fixation_target.set_size(s)
            fixation_target.draw()
            win.flip()

        # fixation period
        for _ in range(n_capture_frames):
            task_vars['background'].draw()
            [a.draw() for a in aruco_ims]
            fixation_target.draw()
            win.flip()

        old_pos = pos

def run_validation(win: visual.Window, config: dict):
    # prepare trial segmentation
    segmenter = SegmentationMarker(win, config["screen"]["refresh_rate"], config["segment_marker"]["duration"],
                                    config["segment_marker"]["size"], config["segment_marker"]["units"], config["segment_marker"]["margin"], config["segment_marker"]["border_bits"], config["segment_marker"]["background_color"])
    # prepare instruction (NB, doesn't respect origin, so set origin as pos)
    textstim = visual.TextStim(win, text="", height=config["instruction_text"]["height"], color=config["instruction_text"]["color"], wrapWidth=9999., pos=config["screen"]["origin"])


    # validation instruction
    textstim.text = 'Follow the fixation target around, and keep fixating the center of it\n\n(Press the spacebar to start)'
    textstim.draw()
    win.flip()
    event.waitKeys()

    # prepare validation
    task_vars = prepare_validation(win, config["validation"], config["screen"])

    # signal validation start
    for m_id in config["validation"]["segment_marker"]["start_IDs"]:
        segmenter.draw(m_id)

    # run validation
    show_validation(win, config["validation"], config["screen"]["refresh_rate"], task_vars)

    # signal validation end
    for m_id in config["validation"]["segment_marker"]["end_IDs"]:
        segmenter.draw(m_id)

def main():
    # read protocol setup
    with open("setup.json") as fp:
        config = json.load(fp)

    try:
        # Open window, check
        mon = monitors.Monitor('temp')
        mon.setWidth(config["screen"]["width"])
        mon.setDistance(config["screen"]["viewing_distance"])
        mon.setSizePix(config["screen"]["resolution"])
        win = visual.Window(monitor=mon,
                            fullscr=True,
                            color=config["screen"]["background_color"],
                            screen=config["screen"]["which_monitor"] or 0,
                            units=config["screen"]["units"],
                            viewPos=config["screen"]["origin"],
                            allowGUI=False,
                            multiSample=True,
                            numSamples=4,
                            infoMsg='')
        win.mouseVisible = False
        if not all((x==y for x,y in zip(win.size,config["screen"]["resolution"]))):
            raise RuntimeError(f'expected resolution of {config["screen"]["resolution"]}, but got {win.size}')
        if 1/win.monitorFramePeriod<config["screen"]["refresh_rate"]-2 or 1/win.monitorFramePeriod>config["screen"]["refresh_rate"]+2:  # check within 2 Hz
            raise RuntimeError(f'expected framerate of {config["screen"]["refresh_rate"]}, but got {1/win.monitorFramePeriod}')

        run_validation(win, config)

    except Exception as e:
        tb_lines = traceback.format_exception(type(e), e, e.__traceback__)
        print("".join(tb_lines))
    finally:
        if 'win' in locals():
            win.close()

if __name__=="__main__":
    main()


"""
Cast raw Viewpointsystem VPS 19 data into common format.

The output directory will contain:
    - frameTimestamps.tsv: frame number and corresponding timestamps for each frame in video
    - worldCamera.mp4: the video from the point-of-view scene camera on the glasses
    - gazeData.tsv: gaze data, where all 2D gaze coordinates are represented w/r/t the world camera,
                    and 3D coordinates in the coordinate frame of the glasses (may need rotation/translation
                    to represent in world camera's coordinate frame)
"""

import shutil
import pathlib
import json
import pandas as pd
import os
import datetime

from ..recording import Recording
from ..eyetracker import EyeTracker
from .. import naming, timestamps, video_utils


def preprocessData(output_dir: str|pathlib.Path=None, source_dir: str|pathlib.Path=None, rec_info: Recording=None, copy_scene_video = True, source_dir_as_relative_path = False, cam_cal_file: str|pathlib.Path=None) -> Recording:
    from . import check_folders, _store_data
    output_dir, source_dir, rec_info, _ = check_folders(output_dir, source_dir, rec_info, EyeTracker.VPS_19)
    print(f'processing: {source_dir.name} -> {output_dir}')


    ### check and copy needed files to the output directory
    print('  Check and copy raw data...')
    ### check tobii recording and get export directory
    if rec_info is not None:
        checkRecording(source_dir, rec_info)
    else:
        rec_info = getRecordingInfo(source_dir)
        if rec_info is None:
            raise RuntimeError(f"The folder {source_dir} is not recognized as a {EyeTracker.VPS_19.value} recording.")

    # make output dir
    if not output_dir.is_dir():
        output_dir.mkdir()


    ### copy the raw data to the output directory
    copyVPS19Recording(source_dir, output_dir, rec_info, copy_scene_video)

    #### prep the copied data...
    print('  Getting camera calibration...')
    if cam_cal_file is not None:
        shutil.copyfile(cam_cal_file, output_dir / naming.scene_camera_calibration_fname)
    else:
        print('    !! No camera calibration provided!')
    print('  Prepping gaze data...')
    gazeDf, frameTimestamps = formatGazeData(source_dir, rec_info)

    _store_data(output_dir, gazeDf, frameTimestamps, rec_info, source_dir_as_relative_path=source_dir_as_relative_path)

    return rec_info


def getRecordingInfo(inputDir: str|pathlib.Path) -> Recording:
    # get recordings. A folder can contain multiple recordings.
    inputDir = pathlib.Path(inputDir)
    # recordings are identified as a tsv and an mkv file with the
    # same name
    recInfos: list[Recording] = []
    for r in inputDir.glob('*.tsv'):
        if not r.with_suffix('.mkv').is_file():
            continue
        recInfos.append(Recording(source_directory=inputDir, eye_tracker=EyeTracker.VPS_19))
        recInfos[-1].name = r.stem
        # get more info
        with open(r,'rt') as f:
            lines = []
            for _ in range(5):
                lines.append(f.readline())
        time_string = lines[2][len('# Recording start: '):].strip()
        recInfos[-1].start_time = timestamps.Timestamp(int(datetime.datetime.fromisoformat(time_string).timestamp()))
        rInfo = json.loads(lines[4][len('# system: '):-1])
        recInfos[-1].glasses_serial = rInfo['glasses']
        recInfos[-1].recording_unit_serial = rInfo['Smart Unit']
        recInfos[-1].firmware_version = rInfo['operating system']

    # should return None if no valid recordings found
    return recInfos if recInfos else None


def checkRecording(inputDir: str|pathlib.Path, recInfo: Recording):
    # check we have an exported gaze data file
    for ext in ('tsv','mkv'):
        file = f'{recInfo.name}.{ext}'
        if not (inputDir / file).is_file():
            if use_return:
                return False
            else:
                raise RuntimeError(f'Recording {recInfo.name} not found: {file} file not found in {inputDir}.')

    return True

def copyVPS19Recording(inputDir: pathlib.Path, outputDir: pathlib.Path, recInfo: Recording, copy_scene_video: bool):
    """
    Copy the relevant files from the specified input dir to the specified output dir
    """
    # Copy relevant files to new directory
    srcFile  = inputDir / f'{recInfo.name}.mkv'

    if copy_scene_video:
        destFile = outputDir / f'{naming.scene_camera_video_fname_stem}.mkv'
        shutil.copy2(srcFile, destFile)
    else:
        destFile = None

    if destFile:
        recInfo.scene_video_file = destFile.name
    else:
        recInfo.scene_video_file =  srcFile.name


def formatGazeData(inputDir: str|pathlib.Path, recInfo: Recording):
    """
    load gazedata json file
    format to get the gaze coordinates w/r/t world camera, and timestamps for every frame of video

    Returns:
        - formatted dataframe with cols for timestamp, frame_idx, and gaze data
        - np array of frame timestamps
    """

    # convert the json file to pandas dataframe
    df = gaze2df(inputDir / f'{recInfo.name}.tsv')

    # read video file, create array of frame timestamps
    frameTimestamps = video_utils.get_frame_timestamps_from_video(recInfo.get_scene_video_path())

    # return the gaze data df and frame time stamps array
    return df, frameTimestamps


def gaze2df(gazeFile: str|pathlib.Path) -> pd.DataFrame:
    """
    convert the .tsv file to a pandas dataframe
    """

    df = pd.read_csv(gazeFile,delimiter='\t',comment='#',index_col=False)

    # get time sync info
    t0 = df['FrontTimeStamp'].iat[0]-df['MediaTimeStamp'].iat[0]

    # rename and reorder columns
    lookup = {'GazeTimeStamp': 'timestamp',
              'MediaFrameIndex': 'frame_idx',
               'Gaze2dX':'gaze_pos_vid_x',
               'Gaze2dY':'gaze_pos_vid_y',
               'PupilCenterLeftX':'gaze_ori_l_x',
               'PupilCenterLeftY':'gaze_ori_l_y',
               'PupilCenterLeftZ':'gaze_ori_l_z',
               'GazeLeftX':'gaze_dir_l_x',
               'GazeLeftY':'gaze_dir_l_y',
               'GazeLeftZ':'gaze_dir_l_z',
               'PupilCenterRightX':'gaze_ori_r_x',
               'PupilCenterRightY':'gaze_ori_r_y',
               'PupilCenterRightZ':'gaze_ori_r_z',
               'GazeRightX':'gaze_dir_r_x',
               'GazeRightY':'gaze_dir_r_y',
               'GazeRightZ':'gaze_dir_r_z',
               'PupilDiaLeft':'pup_diam_l',
               'PupilDiaRight':'pup_diam_r',}
    df=df.rename(columns=lookup)
    # reorder
    idx = [lookup[k] for k in lookup if lookup[k] in df.columns]
    df = df[idx]

    # drop useless gaze frames
    df = df.dropna(axis=0,subset='timestamp')

    # set timestamps t0 to start of video, convert from s to ms and set as index
    df.loc[:,'timestamp'] -= t0
    df.loc[:,'timestamp'] *= 1000.0
    df = df.set_index('timestamp')

    # return the dataframe
    return df


import enum

from . import utils

class EyeTracker(utils.AutoName):
    AdHawk_MindLink = enum.auto()
    Generic         = enum.auto()
    Meta_Aria_Gen_1 = enum.auto()
    Pupil_Core      = enum.auto()
    Pupil_Invisible = enum.auto()
    Pupil_Neon      = enum.auto()
    SeeTrue_STONE   = enum.auto()
    SMI_ETG         = enum.auto()
    Tobii_Glasses_2 = enum.auto()
    Tobii_Glasses_3 = enum.auto()
    VPS_19          = enum.auto()
    Unknown         = enum.auto()
eye_tracker_names = [x.value for x in EyeTracker if x!=EyeTracker.Unknown]

EyeTracker.AdHawk_MindLink.color = utils.hex_to_rgba_0_1("#F0A3FF")
EyeTracker.Generic        .color = utils.hex_to_rgba_0_1("#393939")
EyeTracker.Meta_Aria_Gen_1.color = utils.hex_to_rgba_0_1("#9DCC00")
EyeTracker.Pupil_Core     .color = utils.hex_to_rgba_0_1("#0075DC")
EyeTracker.Pupil_Invisible.color = utils.hex_to_rgba_0_1("#993F00")
EyeTracker.Pupil_Neon     .color = utils.hex_to_rgba_0_1("#4C005C")
EyeTracker.SeeTrue_STONE  .color = utils.hex_to_rgba_0_1("#005C31")
EyeTracker.SMI_ETG        .color = utils.hex_to_rgba_0_1("#2BCE48")
EyeTracker.Tobii_Glasses_2.color = utils.hex_to_rgba_0_1("#FFCC99")
EyeTracker.Tobii_Glasses_3.color = utils.hex_to_rgba_0_1("#94FFB5")
EyeTracker.VPS_19         .color = utils.hex_to_rgba_0_1("#8F7C00")
EyeTracker.Unknown        .color = utils.hex_to_rgba_0_1("#393939")
# other colors left over:
# #C20088, #003380, #FFA405, #FFA8BB, #426600, #FF0010, #5EF1F2, #00998F, #E0FF66, #740AFF, #990000, #FFFF80, #FFE100, #FF5005
# colors taken from P. Green-Armytage (2010). "A Colour Alphabet and the Limits of Colour Coding". Colour: Design & Creativity 5 (10): 1–23.


def string_to_enum(device: str) -> EyeTracker:
    if isinstance(device, EyeTracker):
        return device

    if isinstance(device, str):
        if hasattr(EyeTracker, device):
            return getattr(EyeTracker, device)
        elif device in [e.value for e in EyeTracker]:
            return EyeTracker(device)
        else:
            raise ValueError(f"The string '{device}' is not a known eye tracker type. Known types: {[e.value for e in EyeTracker]}")
    else:
        raise ValueError(f"The variable 'device' should be a string with one of the following values: {[e.value for e in EyeTracker]}")


utils.register_type(utils.CustomTypeEntry(EyeTracker,'__enum.EyeTracker__', utils.enum_val_2_str, lambda x: getattr(EyeTracker, x.split('.')[1])))


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import pathlib

import I2MC

from .. import naming

# assign fixations to targets based on distance
def distance(
        targets: dict[int, np.ndarray],
        fixations: str|pathlib.Path|pd.DataFrame,
        output_directory: str|pathlib.Path,
        do_global_shift = True, max_dist_fac = .5,
        filename_stem: str = naming.fixation_assignment_prefix,
        iteration = 0,
        background_image: tuple[np.ndarray, list[float]] = None,    # (image, extent in mm [l r t b])
        plot_limits: list[list[float]] = None
    ):
    output_directory = pathlib.Path(output_directory)

    # read input if needed
    if not isinstance(fixations,pd.DataFrame):
        fixations = pd.read_csv(fixations,sep='\t',index_col=False)

    # for each target, find closest fixation
    minDur      = 100       # ms
    used        = np.zeros((fixations['start'].size),dtype='bool')
    selected    = np.empty((len(targets),),dtype='int')
    selected[:] = -999

    t_x = np.array([targets[t][0] for t in targets])
    t_y = np.array([targets[t][1] for t in targets])
    off_f_x = off_f_y = off_t_x = off_t_y = 0.
    if do_global_shift:
        # first, center the problem. That means determine and remove any overall shift from the
        # data and the targets, to improve robustness of assigning fixations points to targets.
        # Else, if all data is e.g. shifted up by more than half the distance between
        # validation targets, target assignment would fail
        off_f_x = fixations['xpos'].median()
        off_f_y = fixations['ypos'].median()
        off_t_x = t_x.mean()
        off_t_y = t_y.mean()

    # we furthermore do not assign a fixation to a target if the closest fixation is more than
    # max_dist_fac the intertarget distance away
    # determine intertarget distance, if possible
    dist_lim = np.inf
    if len(t_x)>1:
        # arbitrarily take first target and find closest target to it
        dist = np.hypot(t_x[0]-t_x[1:], t_y[0]-t_y[1:])
        min_dist = dist.min()
        if min_dist > 0:
            dist_lim = min_dist*max_dist_fac

    for i,t in zip(range(len(targets)),targets):
        if np.all(used):
            # all fixations used up, can't assign anything to remaining targets
            continue
        # select fixation
        dist        = np.hypot(fixations['xpos']-off_f_x-(targets[t][0]-off_t_x), fixations['ypos']-off_f_y-(targets[t][1]-off_t_y))
        dist[used]  = np.inf                    # make sure fixations already bound to a target are not used again
        dist[fixations['dur']<minDur] = np.inf  # make sure that fixations that are too short are not selected
        iFix        = np.argmin(dist)
        if dist[iFix]<=dist_lim:
            selected[i] = iFix
            used[iFix]  = True

    # make plot of data overlaid on poster, and show for each target which fixation
    # was selected
    if background_image is not None:
        f       = plt.figure(dpi=300)
        imgplot = plt.imshow(background_image[0],extent=background_image[1],alpha=.5)
        plt.plot(fixations['xpos'],fixations['ypos'],'b-')
        plt.plot(fixations['xpos'],fixations['ypos'],'go')
        if plot_limits is not None:
            plt.xlim(plot_limits[0])
            plt.ylim(plot_limits[1])
        plt.gca().invert_yaxis()
        for i,t in zip(range(len(targets)),targets):
            if selected[i]==-999:
                continue
            plt.plot([fixations['xpos'][selected[i]], targets[t][0]], [fixations['ypos'][selected[i]], targets[t][1]],'r-')

        plt.xlabel('mm')
        plt.ylabel('mm')

        f.savefig(output_directory / f'{filename_stem}_interval_{iteration+1:02d}.png')
        plt.close(f)

    # store selected intervals
    df = pd.DataFrame()
    df.index.name = 'target'
    for i,t in zip(range(len(targets)),targets):
        if selected[i]==-999:
            continue
        df.loc[t,'marker_interval'] = iteration+1
        df.loc[t,'start_timestamp'] = fixations['startT'][selected[i]]
        df.loc[t,  'end_timestamp'] = fixations[  'endT'][selected[i]]

    df.to_csv(output_directory / f'{filename_stem}.tsv',
              mode='w' if iteration==0 else 'a',
              header=iteration==0,
              sep='\t', na_rep='nan', float_format="%.3f")


import pathlib
import importlib.resources
import shutil
import json
import math

from ..config import get_markers, get_targets

def deploy_setup_and_script(output_dir: str|pathlib.Path, overwrite=False) -> list[str]:
    output_dir = pathlib.Path(output_dir)
    if not output_dir.is_dir():
        raise RuntimeError(f'The requested directory "{output_dir}" does not exist')

    # copy over all files
    not_copied: list[str] = []
    for r in ['markerPositions.csv', 'targetPositions.csv', 'setup.json', 'stim.py']:
        out_file = output_dir/r
        if out_file.exists() and not overwrite:
            not_copied.append(r)
            continue
        with importlib.resources.path(__package__, r) as p:
            shutil.copyfile(p, out_file)
    return not_copied

def _get_extent(extent: float, distance: float, psychopy_unit: str) -> float:
    match psychopy_unit:
        case 'cm':
            return extent
        case 'deg' | 'degFlatPos':
            # sizes are not corrected for flat screen when unit is degFlatPos, so same as deg
            # for factor, see comment in _get_position
            return extent * distance*0.017455
        case _:
            raise ValueError(f'PsychoPy unit {psychopy_unit} is not understood')

def _get_position(position: tuple[float,float], distance: float, psychopy_unit: str) -> tuple[float,float]:
    match psychopy_unit:
        case 'cm':
            return position
        case 'deg':
            # 1 deg (centered) at 1 cm is approximately pi/180 (2*tand(.5)), seems to be the logic for PsychoPy
            # this is then apparently hardcoded as 0.017455, which is rounded off a bit wrong...
            fac = distance * 0.017455
            return tuple((p*fac for p in position))
        case 'degFlatPos':
            # positions corrected for flat screen, sizes not
            x,y = (math.radians(x) for x in position)
            return (math.hypot(distance, math.tan(y) * distance) * math.tan(x),
                    math.hypot(distance, math.tan(x) * distance) * math.tan(y))
        case _:
            raise ValueError(f'PsychoPy unit {psychopy_unit} is not understood')

def setup_to_automatic_coding(config_dir: str|pathlib.Path|None, file_name: str='setup.json') -> dict[str, int|list[int]]:
    # load config used to run validation
    # if no config dir specified, load default
    if config_dir is not None:
        with open(pathlib.Path(config_dir)/file_name, 'r') as f:
            setup = json.load(f)
    else:
        with importlib.resources.open_text(__package__, file_name) as f:
            setup = json.load(f)

    out = setup["validation"]["segment_marker"]
    out["border_bits"] = setup["segment_marker"]["border_bits"]
    return out

def setup_to_plane_config(output_dir: str|pathlib.Path, config_dir: str|pathlib.Path|None, file_name: str='setup.json'):
    output_dir = pathlib.Path(output_dir)
    if not output_dir.is_dir():
        raise RuntimeError(f'The requested directory "{output_dir}" does not exist')

    # load config used to run validation
    # if no config dir specified, load default
    if config_dir is not None:
        with open(pathlib.Path(config_dir)/file_name, 'r') as f:
            setup = json.load(f)
    else:
        with importlib.resources.open_text(__package__, file_name) as f:
            setup = json.load(f)

    # get units in which positions are expressed, and viewing distance which may be needed for converting them
    marker_units = setup["validation"]["markers"]["units"] if "units" in setup["validation"]["markers"] else setup["screen"]["units"]
    target_units = setup["validation"]["targets"]["units"] if "units" in setup["validation"]["targets"] else setup["screen"]["units"]
    dist = setup["screen"]["viewing_distance"]
    # load target and marker positions (default ones from package if config dir is not specified)
    target_positions = get_targets(config_dir, setup["validation"]["targets"]["file"], __package__)
    marker_positions = get_markers(config_dir, setup["validation"]["markers"]["file"], __package__)
    # convert positions to cm
    target_positions[['x','y']] = target_positions.apply(lambda r: _get_position((r['x'],r['y']), dist, target_units), axis=1, result_type='expand')
    marker_positions[['x','y']] = marker_positions.apply(lambda r: _get_position((r['x'],r['y']), dist, marker_units), axis=1, result_type='expand')
    # negate y as positive y is upward for PsychoPy script but downward for our logic
    target_positions['y'] = -target_positions['y']
    marker_positions['y'] = -marker_positions['y']
    # Marker and target positions are centers, and relative to center of screen.
    # Our positions should have (0,0) in the top-left. Shift (taking sizes into account so who markers and targets fit on the reference image)
    t_size= _get_extent(setup["validation"]["targets"]["look"]["diameter_max"], dist, target_units)
    m_size= _get_extent(setup["validation"]["markers"]["size"]                , dist, marker_units)
    min_t = target_positions[['x','y']].min()-t_size/2
    min_m = marker_positions[['x','y']].min()-m_size/2
    x_min = min(min_t['x'],min_m['x'])
    y_min = min(min_t['y'],min_m['y'])
    target_positions['x']-=x_min
    marker_positions['x']-=x_min
    target_positions['y']-=y_min
    marker_positions['y']-=y_min

    # get plane size
    max_t = target_positions[['x','y']].max()+t_size/2
    max_m = marker_positions[['x','y']].max()+m_size/2
    x_max = max(max_t['x'],max_m['x'])
    y_max = max(max_t['y'],max_m['y'])

    # write appropriate validationSetup.txt file
    valSetup = {
        'distance': dist,
        'mode': 'cm',
        'markerBorderBits': setup["validation"]["markers"]["border_bits"],
        'markerSide': m_size,
        'markerPosFile': 'markerPositions.csv',
        'targetPosFile': 'targetPositions.csv',
        'targetType': 'Thaler',
        'targetDiameter': t_size,
        'showGrid': 0,
        'gridCols': x_max,
        'gridRows': y_max,
        'minNumMarkers': 3,
        'centerTarget': setup["validation"]["targets"]["center_target"],
        'referencePosterSize': 1920
    }

    # add marker indicator(s) as columns to target file
    for i,_ in enumerate(setup["validation"]["markers"]["replace_IDs"]):
        IDs = target_positions.index.to_numpy()+setup["validation"]["markers"]["replace_ID_start"]+i*setup["validation"]["markers"]["replace_ID_offset"]
        target_positions[f'marker_{i}'] = IDs

    # store everything to files
    with open(output_dir/'validationSetup.txt','w') as f:
        for key,value in valSetup.items():
            f.write(f'{key} = {value}\n')
    target_positions.to_csv(output_dir/valSetup['targetPosFile'], float_format='%.8f')
    marker_positions.to_csv(output_dir/valSetup['markerPosFile'], float_format='%.8f')
